require 'uu/os/cmd/command_handler'

module UU
  module OS
    module CMD

      # Wrapper for result of command. Object of this type is returned
      # in case option :raw_response or :async_invocation is used for
      # {UU::OS::CMD::CommandClient#invoke command invocation}.
      class CommandResult

        # Creates new instance of command result.
        # @param [UU::OS::HTTP::HTTPResponse] http_response Raw command response.
        # @param [Array<#handle_response, #handle_error, #handle_redirect>] handler_chain List of command result handlers.
        # @param [Hash] options Additional options given for request (see {UU::OS::CMD::CommandClient#invoke
        #    CommandClient#invoke} for details.)
        # @private
        def initialize(http_response, handler_chain, options) # :nodoc:
          @http_response = http_response
          @handler_chain = handler_chain
          @options = options
        end

        # Returns result processed by handlers.
        # @return [Object] Command result (actual type depends on used response handlers).
        def get
          process_result
        end

        # Returns unprocessed response of HTTP client. In case of asynchronous
        # command invocation this call is non-blocking, however further manipulation
        # with HTTP response results in waiting for request finish.
        # @return [UU::OS::HTTP::HTTPResponse] Raw HTTP response.
        def unwrap
          @http_response
        end

        private

        # Delegates processing of result to set of response handlers.
        # @return [Object] Command result (actual type depends on used response handlers).
        # @private
        def process_result # :nodoc:
          if (@http_response.status < 300)
            method = :handle_response
          elsif (@http_response.status < 400)
            method = :handle_redirect
          else
            @http_response = ErrorResponse.new(@http_response)
            method = :handle_error
          end
          @handler_chain.each do |handler|
            next if !handler.respond_to?method
            res = handler.send(method, @http_response, @options)
            raise res if res.is_a?Exception
            return res if res != CommandHandler::NEXT
          end
          raise TypeError, "Unable to process result of command using handlers #{@handler_chain}."
        end

        # Mock {UU::OS::HTTP::HTTPResponse} object used to wrap
        # original error response to response with body which
        # can be read multiple times (needed so each handler can
        # read error code and decide whether to handle or not).
        # @private
        class ErrorResponse # :nodoc:

          attr_reader :status

          attr_reader :headers

          attr_reader :body

          attr_reader :request

          def initialize(http_response)
            @status = http_response.status
            @headers = http_response.headers
            @body = http_response.body
            @body = @body.read if @body.respond_to?(:read)
            @request = http_response.request
          end

          def inspect
            "#<#{self.class.name}:0x#{(self.object_id << 1).to_s(16)}>"
          end

        end

      end

    end
  end
end
